@@ -7,14 +7,18 @@ import ipaddress
 import pkgutil
 import os
 
+import six
+
 from cryptoparser.common.base import Serializable
+from cryptoparser.tls.subprotocol import TlsAlertDescription
 from cryptoparser.common.utils import get_leaf_classes
 
-from cryptoparser.tls.subprotocol import TlsAlertDescription
-
+from cryptolyzer.httpx.client import L7ClientHttpBase
+from cryptolyzer.ssh.client import L7ClientSsh
 from cryptolyzer.tls.client import L7ClientTlsBase
 
 
+@six.add_metaclass(abc.ABCMeta)
 class ProtocolHandlerBase(object):
     @classmethod
     def import_plugins(cls):
@@ -50,8 +54,10 @@ class ProtocolHandlerBase(object):
     def get_protocols(cls):
         cls.import_plugins()
 
-        for handler_class in get_leaf_classes(cls):
-            yield handler_class.get_protocol()
+        return sorted([
+            handler_class.get_protocol()
+            for handler_class in get_leaf_classes(cls)
+        ])
 
     @classmethod
     @abc.abstractmethod
@@ -142,3 +148,53 @@ class AnalyzerTlsBase(object):
     @abc.abstractmethod
     def analyze(self, analyzable, protocol_version):
         raise NotImplementedError()
+
+
+class ProtocolHandlerTlsBase(ProtocolHandlerBase):
+    @classmethod
+    @abc.abstractmethod
+    def get_protocol_version(cls):
+        raise NotImplementedError()
+
+    @classmethod
+    def get_protocol(cls):
+        return cls.get_protocol_version().identifier
+
+    @classmethod
+    def _get_analyzer_args(cls):
+        return ([], {'protocol_version': cls.get_protocol_version()})
+
+
+class ProtocolHandlerTlsExactVersion(ProtocolHandlerTlsBase):
+    @classmethod
+    @abc.abstractmethod
+    def get_protocol_version(cls):
+        raise NotImplementedError()
+
+
+class AnalyzerSshBase(object):
+    @classmethod
+    def get_clients(cls):
+        return list(get_leaf_classes(L7ClientSsh))
+
+    @classmethod
+    def get_default_scheme(cls):
+        return 'ssh'
+
+    @abc.abstractmethod
+    def analyze(self, analyzable):
+        raise NotImplementedError()
+
+
+class AnalyzerHttpBase(object):
+    @classmethod
+    def get_clients(cls):
+        return list(get_leaf_classes(L7ClientHttpBase))
+
+    @classmethod
+    def get_default_scheme(cls):
+        return 'http'
+
+    @abc.abstractmethod
+    def analyze(self, analyzable, protocol_version):
+        raise NotImplementedError()
